/**********************************************************************
 * queue_func.c                                            October 1999
 * Horms                                             horms@vergenet.net
 *
 * Functions build around the queue ADT
 *
 * Licenced under the GNU GPL
 *
 **********************************************************************/

#include "queue_func.h"
#include "options.h"


/**********************************************************************
 * read_line
 * read a line from fd and parse it into a queue of tokens
 * line is read by making repeated calls to read_token
 * pre: fd: file descriptor to read from
 *      buf: buffer to store bytes read from server in
 *      n: pointer to size_t containing the size of literal_buf
 * post: Token is read from fd into token
 *       If literal_buf is not NULL, and n is not NULL and *n is not 0
 *       Bytes read from fd are copied to literal_buf.
 * return: token
 *         NULL on error
 * Note: If buf is being filled and space is exausted function will
 *       return what has been read so far. (No buffer overflows today)
 **********************************************************************/

queue *read_line(const int fd, unsigned char *buf, size_t *n){
  token_t *t=NULL;
  size_t buf_offset=0;
  size_t buf_remaining;
  queue *q;
  int do_literal;

  extern options_t opt;

  if(buf!=NULL && n!=NULL && *n!=0){
    do_literal=1;
    buf_remaining=*n;
  }
  else{
    buf=NULL;
    buf_remaining=0;
    do_literal=0;
  }

  if((q=create_queue())==NULL){
    if(opt.debug){
      syslog(LOG_WARNING, "read_line: create_queue");
    }
    return(NULL);
  }
 
  do{
    if((t=read_token(fd,(buf==NULL)?NULL:buf+buf_offset,&buf_remaining))==NULL){
      if(opt.debug){
        syslog(LOG_WARNING, "read_line: read_token");
      }
      destroy_queue(q, DESTROY_TOKEN);
      return(NULL);
    }

    if(do_literal){
      buf_offset+=buf_remaining;
      buf_remaining=*n-buf_offset;
    }

    if((q=push(q, (perdition_queue_t)t, DESTROY_TOKEN))==NULL){
      if(opt.debug){
        syslog(LOG_WARNING, "read_line: push");
      }
      return(NULL);
    }
  }while(!token_is_eol(t) && !(do_literal && buf_offset>=*n));

  if(do_literal){
    *n=buf_offset;
  }

  return(q);
}


/**********************************************************************
 * perdition_queue_to_string
 * convert the contents of a queue of tokens into a string
 * a space ( ) is inserted in the resultant string between each
 * token
 * pre: q queue to dump as a string
 * post: a string is allocated and the quie is dumped to the string
 *       the string is '\0' terminated
 * return: allocated string
 *         NULL on error
 **********************************************************************/

char *perdition_queue_to_string(queue *q){
  token_t *t;
  queue *stack=NULL;
  size_t length=0;
  char *string;
  char *pos;

  extern options_t opt;
  
  if((stack=create_queue())==NULL){
    if(opt.debug){
      syslog(LOG_WARNING, "perdition_queue_to_string: create_queue");
    }
    return(NULL);
  }

  while(q->first!=NULL){
    if((q=pop(q, (perdition_queue_t *)&t))==NULL){
      if(opt.debug){
        syslog(LOG_WARNING, "perdition_queue_to_string: pop 1");
      }
      return(NULL);
    }

    length+=1+t->n;

    if((stack=push(stack, (perdition_queue_t)t, DESTROY_TOKEN))==NULL){
      if(opt.debug){
        syslog(LOG_WARNING, "perdition_queue_to_string: push");
      }
      return(NULL);
    }
  }

  destroy_queue(q, DESTROY_TOKEN);

  if((string=(char*)malloc(sizeof(char)*length))==NULL){
    if(opt.debug){
      syslog(LOG_WARNING, "queue_string: malloc");
    }
    destroy_queue(stack, DESTROY_TOKEN);
    return(NULL);
  }

  pos=string;
  while(stack->first!=NULL){
    if((stack=pop(stack, (perdition_queue_t *)&t))==NULL){
      if(opt.debug){
        syslog(LOG_WARNING, "perdition_queue_to_string: pop 2");
      }
      free(string);
      return(NULL);
    }
    
    strncpy(pos, t->buf, t->n);
    pos+=t->n;
    *pos++=' ';
  }
  
  destroy_queue(stack, DESTROY_TOKEN);
  *--pos='\0';

  return(string);
}
